#!/usr/bin/python3

from __future__ import annotations
from struct import pack
from tinytag import tinytag

from .. import SmallTag

if False:
    from collections.abc import Callable, Iterator  # noqa
    from typing import BinaryIO


class _ID3(SmallTag, tinytag._ID3):
    @staticmethod
    def _synchsafe(unsynchsafe_int: int) -> tuple[int, ...]:
        ints = (
            (unsynchsafe_int >> 21) & 0x7F,
            (unsynchsafe_int >> 14) & 0x7F,
            (unsynchsafe_int >> 7) & 0x7F,
            unsynchsafe_int & 0x7F
        )

        return ints

    def write(self, advanced: dict[str, str | float | list[str]] | None=None, file_obj: BinaryIO=None) -> None:
        should_close_file = self._set_filehandler_for_writing(file_obj)

        new_tag = self._compose_id3v2_tag(advanced)

        size, extended, major = self._parse_id3v2_header(self._filehandler)

        file_content = self._filehandler.read()
        audio = file_content[size:]

        self._filehandler.seek(0)
        self._filehandler.write(new_tag + audio)
        self._filehandler.truncate()

        if should_close_file:
            self._filehandler.close()

    def compose_tag(self, advanced: dict | None=None) -> bytes:
        return self._compose_id3v2_tag(advanced)

    def _compose_id3v2_tag(self, advanced: dict | None=None) -> bytes:
        # change keys to values in the mapping dict
        self._ID3_WRITE_MAPPING = {}  # noqa
        for frame_id, name in self._ID3_MAPPING.items():
            if len(frame_id) == 4:
                self._ID3_WRITE_MAPPING[name] = frame_id

        frames = self._compose_id3v2_frames(advanced)
        frame_size = len(frames)

        header = self._compose_id3v2_header(frame_size)

        tag = header + frames

        return tag

    def _compose_id3v2_header(self, size: int) -> bytes:
        header = b"ID3\x04\x00\x00"

        synchsafe_size = self._synchsafe(size)
        header += pack("4B", *synchsafe_size)

        return header

    def _compose_id3v2_frames(self, advanced: dict | None=None) -> bytes:
        if advanced is None:
            tag_dict = self.as_dict()

        else:
            tag_dict = advanced

        frames = b""

        for field_name, field_value in tag_dict.items():
            if field_name not in self._ID3_WRITE_MAPPING:
                other_field = "other." + field_name

                if other_field in self._ID3_WRITE_MAPPING:
                    frames += self._compose_id3v2_frame(other_field, field_value)

                continue

            frames += self._compose_id3v2_frame(field_name, field_value)

        return frames

    def _compose_id3v2_frame(self, field_name, field_value) -> bytes:
        frame_id = bytes(self._ID3_WRITE_MAPPING[field_name], "ISO-8859-1")

        if isinstance(field_value, list):
            field_value = field_value[0]

        if isinstance(field_value, str):
            frame_value = bytes(field_value, "ISO-8859-1")

            # set format byte to ISO-8859-1 on normal text frames
            # (link-frames are always ISO-8859-1 formatted and don't need the format byte)
            if frame_id.startswith(b"T"):
                frame_value = b"\x00" + frame_value

        elif isinstance(field_value, int) and frame_id.startswith(b"T"):
            frame_value = b"\x00" + bytes(str(field_value), "ISO-8859-1")

        else:
            print(f"Writing of '{field_name}' (type {type(field_value)}) is not implemented.")
            return b""

        frame_size = pack("4B", *self._synchsafe(len(frame_value)))

        frame_data = frame_id + frame_size + b"\x00\x00" + frame_value

        return frame_data
